use std::{
    collections::BTreeMap,
    fs::File,
    io::{BufRead, BufReader, Write},
    path::PathBuf,
    str::FromStr,
};

use anyhow::{anyhow, Context, Result};
use chrono::{DateTime, Utc};
use log::{debug, info, trace};
use regex::Regex;
use semver::Version;

use crate::{Position, PositionLayer, PsbId, PsbQaqcResult};

/// QAQC results for each boards extracted from master log.
///
/// Temporary stores raw values.
/// TODO: specify each filed types.(maybe multi values)
#[derive(Debug)]
pub struct MasterBoardResult {
    pub id: PsbId,
    pub qspip: u8,
    pub recov: u8,
    pub power: u8,
    pub clock: u8,
    pub asdtp: u8,
    pub reset: u16,
    pub result: u8,
}

impl MasterBoardResult {
    /// Check the board needs to be retested with reason.
    ///
    /// None if it doesn't need retest.
    pub fn is_need_retest(&self) -> Option<String> {
        let mut result = false;
        let mut reasons = vec![];
        if self.qspip != 1 {
            result = true;
            reasons.push(format!("QSPIp is not 1({})", self.qspip));
        }
        if self.recov != 1 {
            result = true;
            reasons.push(format!("Recov is not 1({})", self.recov));
        }
        if self.power != 1 {
            result = true;
            reasons.push(format!("Power is not 1({})", self.qspip));
        }
        if self.clock != 1 {
            result = true;
            reasons.push(format!("Clock is not 1({})", self.clock));
        }
        match result {
            true => Some(reasons.join(", ")),
            false => None,
        }
    }
}

/// Full result for a single QAQC run from a log file on JATHub master.
#[derive(Debug)]
pub struct MasterLogResult {
    pub version: Version,
    pub datetime: DateTime<Utc>,
    pub runid: u32,
    pub shifter: String,
    pub board_results: BTreeMap<Position, MasterBoardResult>,
    pub filename: String,
}

/// Get version of shift script.
fn extract_version(line: &str) -> Result<Version> {
    Ok(line
        .split_whitespace()
        .nth(2)
        .context("Invalid log format(version)")?
        .parse()?)
}

/// Get runid.
fn extract_runid_line(line: &str) -> Result<u32> {
    let re = Regex::new(r"^QAQC runid: (\d+)$").unwrap();
    let caps = re.captures(line).unwrap();
    trace!("Regex {:?}", caps);
    let runid = u32::from_str(caps.get(1).unwrap().into())?;
    Ok(runid)
}

/// Get shifters from shifter line
fn extract_shifter_line(line: &str) -> Result<String> {
    let re = Regex::new(r"^Shifters: (.+)$").unwrap();
    let caps = re.captures(line).unwrap();
    trace!("Regex {:?}", caps);
    Ok(caps.get(1).unwrap().as_str().to_owned())
}

/// Get position and PSB ID pair from a master log line.
fn extract_position_id(line: &str) -> Result<(Position, PsbId)> {
    let re = Regex::new(r"Position / assigned-ID : (.+) / (.+)").unwrap();
    let caps = re.captures(line).context("No capture")?;
    let pos = Position::from_str(caps.get(1).unwrap().into())?;
    let id = PsbId::from_str(caps.get(2).unwrap().into())?;
    Ok((pos, id))
}

impl MasterLogResult {
    /// Parse log file on master jathub.
    pub fn parse_file(master_log: PathBuf) -> Result<MasterLogResult> {
        let file = File::open(master_log.clone())?;
        let reader = BufReader::new(file);
        MasterLogResult::parse_file_with_filename(
            reader,
            master_log
                .file_name()
                .unwrap()
                .to_str()
                .unwrap()
                .to_string(),
        )
    }

    fn parse_file_with_filename(file: impl BufRead, filename: String) -> Result<MasterLogResult> {
        let mut lines = file.lines();

        let version = {
            let line = lines.next().context("Invalid log format(no versions)")??;
            extract_version(&line)?
        };

        let _sep = lines.next().context("Invalid log format(separator)")??;

        let datetime: DateTime<Utc> = {
            let line = lines
                .next()
                .context("Invalid log format(no datetime line)")??;
            DateTime::parse_from_str(&line, "Date: %+")
                .context("Invalid datetime format (must be ISO 8601)")?
                .to_utc()
        };

        let runid: u32 = {
            let line = lines
                .next()
                .context("Invalid log format(no runid line)")??;
            extract_runid_line(&line)?
        };

        let shifter = {
            let line = lines
                .next()
                .context("Invalid log format(no shifter line)")??;
            extract_shifter_line(&line)?
        };

        let _sep = lines.next().context("Invalid log format")?;

        if !lines
            .next()
            .context("Invalid log format")??
            .starts_with("PBS Assignment:")
        {
            return Err(anyhow!("Invalid log format"));
        }

        let mut assignments = BTreeMap::new();
        // till 19 for `===========`
        for i in 0..19 {
            let line = lines.next().context("Unexpected EOF")??;
            if line.starts_with("=========") {
                debug!("End of assignments");
                break;
            }
            debug!("line: {}", line);
            let (pos, id) = extract_position_id(&line)?;
            trace!("pos: {}", pos);
            match assignments.insert(pos.clone(), id) {
                None => (),
                Some(old_id) => {
                    return Err(anyhow!(
                        "Value already exists on row {}: {:?} => {:?}",
                        i,
                        pos,
                        old_id
                    ));
                }
            };
        }
        trace!("Read all PBS assignments");
        info!("{:?}", assignments);

        // TODO: stricter validation for header?
        if !lines
            .next()
            .context("Invalid log format")??
            .contains("QAQC status")
        {
            info!("{:?}", lines.next());
            return Err(anyhow!("Invalid log format(result table header)"));
        }
        let _sep = lines
            .next()
            .context("Invalid log format(result table separator)")??;

        if !lines
            .next()
            .context("Invalid log format")??
            .contains("Station0")
        {
            return Err(anyhow!("Invalid log format(result Station0)"));
        }

        let mut board_results = BTreeMap::new();
        for station_minor in [0, 1] {
            info!("Result for {:?}", station_minor);

            // Loop for sufficiently large number to reach `Station1`
            for _ in 1..20 {
                let line = lines.next().context("Invalid log format(result body)")??;
                trace!("line: {}", line);
                if line.contains("======") {
                    // end of status
                    break;
                }
                if line.contains(&format!("Station{}", station_minor + 1)) {
                    debug!("Next station: line: {}", line);
                    break;
                }
                let parts: Vec<&str> = line.split('|').collect();
                let first = parts.first().context("No col 1")?;
                let raw_station_id = {
                    let re = Regex::new(r"JATHub_( ?\d*)$")?;
                    re.captures(first).map(|v| {
                        v.get(1)
                            .unwrap()
                            .as_str()
                            .split_whitespace()
                            .next()
                            .unwrap()
                            .parse::<u8>()
                    })
                }
                .context(format!("Invalid station format: {}", line))??;
                let station_id = match station_minor {
                    0 => raw_station_id,
                    1 => raw_station_id - 10,
                    _ => panic!("Unexpected"),
                };
                trace!("Row {} {:?}", station_id, parts);
                if parts.len() != 9 {
                    return Err(anyhow!(
                        "Invalid number of results(expected: 9, detected: {})",
                        parts.len()
                    ));
                }

                // Origin is different (-1)
                let pos = Position {
                    major: PositionLayer::B,
                    minor: station_minor,
                    patch: station_id,
                };
                debug!("pos from table {}", pos);

                let psbid = assignments
                    .get(&pos)
                    .context(format!("No board on pos {}", pos))?
                    .clone();

                let result = MasterBoardResult {
                    id: psbid,
                    qspip: parts
                        .get(1)
                        .map(|v| v.split_whitespace().next().unwrap().parse())
                        .context("Invalid qspip")??,
                    recov: parts
                        .get(2)
                        .map(|v| v.split_whitespace().next().unwrap().parse())
                        .context("Invalid recov")??,
                    power: parts
                        .get(3)
                        .map(|v| v.split_whitespace().next().unwrap().parse())
                        .context("Invalid power")??,
                    clock: parts
                        .get(4)
                        .map(|v| v.split_whitespace().next().unwrap().parse())
                        .context("Invalid clock")??,
                    asdtp: parts
                        .get(5)
                        .map(|v| v.split_whitespace().next().unwrap().parse())
                        .context("Invalid asdtp")??,
                    reset: parts
                        .get(6)
                        .map(|v| v.split_whitespace().next().unwrap().parse())
                        .context("Invalid reset")??,
                    result: parts
                        .get(7)
                        .map(|v| v.split_whitespace().next().unwrap().parse())
                        .context("Invalid result")??,
                };

                match board_results.insert(pos, result) {
                    None => (),
                    Some(v) => {
                        panic!("Unexpected value already exists: {:?}", v)
                    }
                }
            }
        }

        debug!("{:#?}", board_results);

        Ok(MasterLogResult {
            version,
            datetime,
            runid,
            shifter,
            board_results,
            filename,
        })
    }

    pub fn print_boards_to_retest(&self, mut wtr: impl Write) -> Result<()> {
        for (pos, boardresult) in &self.board_results {
            if let Some(reason) = boardresult.is_need_retest() {
                writeln!(
                    wtr,
                    "Board {} at {} need retest for {}",
                    boardresult.id.id, pos, reason
                )?;
            }
        }
        Ok(())
    }
}

pub fn get_output_filename(result: &MasterLogResult) -> PathBuf {
    PathBuf::from(format!("out_{}.csv", result.runid))
}

#[cfg(test)]
mod test {
    use semver::Version;

    use crate::masterlog::extract_runid_line;

    use super::{Position, PositionLayer, PsbId};

    use super::{extract_position_id, extract_shifter_line, extract_version};

    #[test]
    fn parse_version() {
        assert_eq!(
            extract_version("Shift script: 1.0.0").unwrap(),
            Version::new(1, 0, 0)
        )
    }

    #[test]
    fn parse_runid_line() {
        assert_eq!(extract_runid_line("QAQC runid: 7").unwrap(), 7);
        assert_eq!(extract_runid_line("QAQC runid: 12345").unwrap(), 12345);
    }

    #[test]
    fn parse_shifter_line() {
        assert_eq!(extract_shifter_line("Shifters: foo").unwrap(), "foo");
        assert_eq!(
            extract_shifter_line("Shifters: foo bar").unwrap(),
            "foo bar"
        );
    }

    #[test]
    fn parse_pos_id_line() {
        assert_eq!(
            extract_position_id("Position / assigned-ID : A-0-0 / PS0004").unwrap(),
            (
                Position {
                    major: PositionLayer::A,
                    minor: 0,
                    patch: 0,
                },
                PsbId::new(4)
            )
        );
        assert_eq!(
            extract_position_id("Position / assigned-ID : A-1-7 / PS0108").unwrap(),
            (
                Position {
                    major: PositionLayer::A,
                    minor: 1,
                    patch: 7,
                },
                PsbId::new(108)
            )
        );
        assert_ne!(
            extract_position_id("Position / assigned-ID : A-1-7 / PS0108").unwrap(),
            (
                Position {
                    major: PositionLayer::A,
                    minor: 0,
                    patch: 7,
                },
                PsbId::new(106)
            )
        );
    }
}
